from sys import maxsize
from typing import Callable, Type, Union, _Final
from inspect import cleandoc, signature
from functools import wraps


def sanitize_docs(docs):
    return cleandoc(docs).replace('\n', ' ')


def enclose(strings, char):
    return [f"{char}{s}{char}" for s in strings]


def parametrized_decorator_method(decorator):
    """
    Meta decorator for cleanly creating decorators that accept parameters.
    Adapted from https://stackoverflow.com/a/26151604/ to use on methods.
    """

    @wraps(decorator)
    def layer(self, *args, **kwargs):
        def apply(f):
            return decorator(self, f, *args, **kwargs)
        return apply
    return layer


def get_arity(f: Callable):
    return len(signature(f).parameters)


def last_argument_is_of_type(f: Callable, t: Type):
    sig = signature(f)
    params = list(sig.parameters.values())
    if not params:
        return False
    return params[-1].annotation is t


def call_with_typed_args(f, *args, optional_args=False):
    sig = signature(f)
    coerced = []
    for param, arg in zip(sig.parameters.values(), args):
        optional = optional_args
        if param.annotation is not sig.empty:
            t = param.annotation
            if hasattr(t, '__origin__') and t.__origin__ is Union and t.__args__[-1] is type(None):
                # argument is of type `Optional[type]`
                t = t.__args__[0]
                optional = True
            try:
                arg = t(arg)
            except TypeError as e:
                if not isinstance(t, _Final):
                    # the type could be instantiated
                    arg = None
                # used a type from `typing`, which are ignored, ex: `AnyStr`
            except ValueError as e:
                if arg:
                    # invalid arg, ex: `int('1.0')`
                    raise e
                if t in (int, float):
                    arg = maxsize
        if arg == '' or arg == maxsize:
            if not optional:
                raise ValueError
            if param.default is not sig.empty:
                arg = param.default
            elif arg == maxsize:
                arg = 0
        coerced.append(arg)
    return f(*coerced)



import logging
from time import sleep
from dgg_chat import DGGChat

from .exceptions import InvalidMessageError, AnonymousConnectionError
from ._commands import Commands, Message, Whisper
from ._utils import parametrized_decorator_method, enclose


class DGGChatBot:
    def __init__(self, auth_token=None, command_prefix='!', extra_help=''):
        """
        Parameters
        ----------
        `auth_token` : `str`

            an authentication token for a dgg account.
            can be created at `https://www.destiny.gg/profile/developer`.       

        `command_prefix` : `str`

            the character sequence at the start of a message
            to indicate it's a command.
            `"!"` by default.

        `extra_help` : `str`

            extra details to be put at the end of the default help function reply.
        """
        self._auth_token = auth_token
        self.command_prefix = command_prefix
        self.extra_help = extra_help
        self._last_whisper_received = None

        self._commands = Commands(command_prefix)

        self._chat = DGGChat(auth_token)

        self._setup_default_commands()
        self._setup_chat_handlers()

    @property
    def auth_token(self):
        return self._auth_token

    @property
    def commands(self):
        return self._commands

    @property
    def chat(self):
        return self._chat

    def _setup_default_commands(self):
        @self.on_command('help', 'h', 'commands', optional_args=True)
        def on_help(cmd):
            """documentation set below"""

            if cmd:
                if not (command := self._commands.get_root(cmd)):
                    return self.reply('The command `{cmd}` was not found.')
                aliases = ', '.join(enclose(command.aliases, '"')) or 'none'
                desc = command.handler.description
                msg = f'"{self.command_prefix}{command.keyword}" (aliases: {aliases}) -> {desc}'
                return self.reply(msg)

            commands = ', '.join(enclose(self._commands.all_aliases, '"'))
            if not commands:
                msg = 'No commands are available ☹.'
                return self.reply(msg)
            msg = f'Available commands: {commands}. For more info about a specific command, try "{self.command_prefix}help <command>". {self.extra_help}'
            return self.reply(msg)

        # setting manually since formatting is needed
        on_help.__doc__ = f"""
        The command you're using! 
        Use it to get info about available commands.
        Examples: "{self.command_prefix}help", "{self.command_prefix}help <command>".
        """

        @self.on_unknown_command
        def on_unknown_command(command):
            msg = f"""Sorry, I don't know that one 🤔. Try "{self.command_prefix}help"."""
            self.reply(msg)

        @self.on_regular_message
        def on_regular_message(message: Message):
            msg = f"""Hey, {message.user.nick}! I'm a bot 🤖 *beep* *boop*. To check what I can do, try "{self.command_prefix}help"."""
            self.reply(msg)

        @self.on_invalid_arguments
        def on_invalid_arguments(command, error_message=''):
            keyword = command.keyword
            error_message = f" Error message: {error_message}." if error_message else ''
            msg = f"""I don't think this you how you use that 🤔.{error_message} Try "{self.command_prefix}help {keyword}" for more info."""
            self.reply(msg)

        @self.on_fail
        def on_fail(command, error_message=''):
            error_message = f" Error message: {error_message}." if error_message else ''
            msg = f"""Something went wrong while processing your command.{error_message}"""
            self.reply(msg)

    def _setup_chat_handlers(self):
        # @self._chat.on_chat_message
        @self._chat.on_whisper
        def on_whisper(whisper: Whisper):
            self._last_whisper_received = whisper
            self._commands.handle(whisper)

    def connect(self):
        self._chat.connect()

    def disconnect(self):
        self._chat.disconnect()

    def run_forever(self):
        self._chat.run_forever()

    def on_regular_message(self, f):
        self._commands.on_regular_message = f
        return f

    def on_unknown_command(self, f):
        self._commands.on_unknown_command = f
        return f

    def on_invalid_arguments(self, f):
        self._commands.on_invalid_arguments = f
        return f

    def on_fail(self, f):
        self._commands.on_fail = f
        return f

    def before_every_command(self, f):
        return self._commands.add_before_every_command(f)

    def after_every_command(self, f):
        return self._commands.add_after_every_command(f)

    @parametrized_decorator_method
    def on_command(
        self, f, keyword, *aliases,
        override=False,
        optional_args=False,
    ):
        if not self._auth_token:
            raise AnonymousConnectionError(
                'setting command handlers is not very useful without a connection'
            )

        self._commands.add(
            f, keyword, *aliases,
            override=override,
            optional_args=optional_args
        )
        return f

    def reply_multiline(self, messages):
        """
        Replies to the last whisper received with multiple messages.
        Use sparingly, since you're likely to get throttled.
        """
        for m in messages:
            self.reply(m)

    def reply(self, message):
        """
        Replies to the last whisper received.
        """
        if not self._chat.message_is_valid(message):
            raise InvalidMessageError(message)
        user = self._last_whisper_received.user.nick
        self._chat.send_whisper(user, message)



class CommandDescriptionTooLongError(Exception):
    """
    Raised when the command description (via `f.__doc__`) is too long.
    """


class DuplicateCommandError(Exception):
    """
    Raised when a keyword or alias has already been used for a command.
    """

    def __init__(self, alias):
        msg = f"`{alias}` was already used as a keyword or alias for a command. use `override=True` to override it"
        super().__init__(msg)


class InvalidCommandError(Exception):
    """
    Raised when the message does not start with command prefix.
    """


class UnknownCommandError(Exception):
    """
    Raised when the command is not recognized.
    """


class InvalidCommandArgumentsError(Exception):
    """
    Raised when the arguments used for the command were invalid.
    """



from dgg_chat.exceptions import InvalidMessageError, AnonymousConnectionError
from ._exceptions import (
    CommandDescriptionTooLongError,
    DuplicateCommandError,
    InvalidCommandError,
    InvalidCommandArgumentsError,
    UnknownCommandError,
)



import logging
from typing import Union, Callable, Dict, Set, AnyStr
from dgg_chat.messages import ChatMessage, Whisper, EventTypes

from .exceptions import (
    CommandDescriptionTooLongError,
    DuplicateCommandError,
    InvalidMessageError,
    InvalidCommandError,
    InvalidCommandArgumentsError,
    UnknownCommandError,
)
from ._utils import (
    get_arity,
    sanitize_docs,
    call_with_typed_args,
    last_argument_is_of_type,
)


Message = Union[ChatMessage, Whisper]


class Handler:
    MAX_DESCRIPTION_LENGTH = 400

    def __init__(self, f: Callable, optional_args=False):
        self._f = f
        self.description = f.__doc__ or 'No description.'
        self.optional_args = optional_args

        self.description = sanitize_docs(self.description)
        if len(self.description) > self.MAX_DESCRIPTION_LENGTH:
            raise CommandDescriptionTooLongError(self.description)

    def __call__(self, raw_message: Message, *args):
        exp_raw = self.expects_raw_message
        actual_arity = self.arity - exp_raw

        if not actual_arity:
            if exp_raw:
                return self._f(raw_message)
            return self._f()

        args = list(args)

        if len(args) < actual_arity:
            # add empty strings to args to match the handler arity
            diff = actual_arity - len(args)
            args.extend('' for _ in range(diff))

        if len(args) > actual_arity:
            # extra arguments are clumped as the last one
            # (second to last if expects raw message)
            split = actual_arity - 1
            args, remaining = args[:split], ' '.join(args[split:])
            args.append(remaining)

        if exp_raw:
            args.append(raw_message)

        try:
            return call_with_typed_args(
                self._f, *args, optional_args=self.optional_args
            )
        except ValueError:
            raise InvalidCommandArgumentsError('invalid arguments')

    @property
    def arity(self):
        return get_arity(self._f)

    @property
    def expects_raw_message(self):
        return last_argument_is_of_type(self._f, Message)


class Command:
    def __init__(self, handler: Handler, keyword, *aliases):
        self._handler = handler
        self._keyword = keyword.lower()
        self._aliases = set(a.lower() for a in aliases)

    def __repr__(self):
        return f"Command(keyword='{self._keyword}', aliases={self._aliases})"

    @property
    def keyword(self):
        return self._keyword

    @property
    def aliases(self):
        return self._aliases

    @property
    def handler(self):
        return self._handler

    def is_alias(self, alias):
        return alias in self._aliases

    def add_alias(self, alias):
        self._aliases.add(alias)


class Commands:
    def __init__(self, command_prefix):
        self.command_prefix = command_prefix
        self._commands: Dict[Keyword, Command] = {}
        self._on_regular_message: Callable = None
        self._on_unknown_command: Callable = None
        self._on_invalid_arguments: Callable = None
        self._on_fail: Callable = None
        self._before_every_command: Set[Callable] = set()
        self._after_every_command: Set[Callable] = set()

    @property
    def on_regular_message(self):
        return self._on_regular_message

    @on_regular_message.setter
    def on_regular_message(self, f):
        # just in case restrictions are needed in the future
        self._on_regular_message = f

    @property
    def on_unknown_command(self):
        return self._on_unknown_command

    @on_unknown_command.setter
    def on_unknown_command(self, f):
        # just in case restrictions are needed in the future
        self._on_unknown_command = f

    @property
    def on_invalid_arguments(self):
        return self._on_invalid_arguments

    @on_invalid_arguments.setter
    def on_invalid_arguments(self, f):
        # just in case restrictions are needed in the future
        self._on_invalid_arguments = f

    @property
    def on_fail(self):
        return self._on_fail

    @on_fail.setter
    def on_fail(self, f):
        # just in case restrictions are needed in the future
        self._on_fail = f

    @property
    def all_aliases(self):
        aliases = []
        for keyword, command in self._commands.items():
            aliases.append(keyword)
            aliases.extend(command.aliases)
        return aliases

    def get_root(self, alias):
        if command := self._commands.get(alias):
            return command
        for command in self._commands.values():
            if command.is_alias(alias):
                return command

    def add_before_every_command(self, f):
        self._before_every_command.add(f)
        return f

    def add_after_every_command(self, f):
        self._after_every_command.add(f)
        return f

    def add(
        self, f, keyword, *aliases,
        override=False,
        optional_args=False,
    ):
        for alias in aliases:
            if self.get_root(alias):
                raise DuplicateCommandError(alias)

        if self.get_root(keyword) and not override:
            raise DuplicateCommandError(keyword)

        handler = Handler(f, optional_args)
        command = Command(handler, keyword, *aliases)
        self._commands[keyword] = command

    def handle(self, message: Message):
        command = message.content.strip()

        if not command.startswith(self.command_prefix):
            if self._on_regular_message:
                return self._on_regular_message(message)
            raise InvalidCommandError(command)

        keyword, *args = command.lstrip(self.command_prefix).strip().split(' ')

        if not (command := self.get_root(keyword.lower())):
            if self._on_unknown_command:
                return self._on_unknown_command(keyword)
            raise UnknownCommandError(command)

        exc = None
        try:
            try:
                for h in self._before_every_command:
                    h(message, command.keyword, *args)
                command.handler(message, *args)
            except InvalidCommandArgumentsError as e:
                logging.debug(f"invalid arguments: {command}, {args}. Error: {e}")
                self._on_invalid_arguments(command, str(e))
                raise e
            except Exception as e:
                logging.debug(f"something went wrong: {command}, {args}. Error: {e}")
                self._on_fail(command, str(e))
                raise e
        except Exception as e:
            exc = e
        finally:
            for h in self._after_every_command:
                h(message, exc, command.keyword, *args)



from setuptools import setup, find_packages
from pathlib import Path

from dgg_chat_bot import __version__

long_description = Path('README.md').read_text()

setup(
    name='dgg-chat-bot',
    packages=find_packages(),
    version=__version__,
    license='MIT',
    description='Build chat bots for the destiny.gg chat',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Gabriel Jablonski',
    author_email='contact@gabrieljablonski.com',
    url='https://github.com/gabrieljablonski/dgg-chat-bot',
    keywords=['chat-bot', 'chat', 'destinygg', 'dgg'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.8',
    install_requires=[
        'dgg_chat',
    ],
)
